searchNodes=[{"type":"module","title":"Toolbox.Incident","doc":"Module extends Toolbox.Workflow and abstracts how regular incident behaves. Wraps around Toolbox.Workflow and adds some additional callbacks to manage incident in asset map. This module works very much like regular workflow, but some additional properties can be specified. Contrary to a regular workflow, this automatically generates OAs to manage the incident in asset map and therefore syncs the general state of this workflow with the incident asset. Start by creating a definition (see new/0 , add_transition/2 and build/1 ) which describes the workflow of the incident. Then, you can create a new instance based on that definition with new_instance/7 .","ref":"Toolbox.Incident.html"},{"type":"function","title":"Toolbox.Incident.add_transition/2","doc":"Adds a new transition to incident workflow definition. Incident workflow is a finite state machine. Transition defines how workflow changes its state from one to another and what must be done during the state change. Incident transition is defined by params parameter. params is a keyword list where each transition parameter has its key. Parameters may be either required or optional. Some parameters have predefined keys. These parameters are used to control workflow. The developer of the scenario can add his own parameters to enrich function of the workflow with features that the standard workflow module does not provide. These parameters can have any key except predefined ones. If add_transition/2 succeeds it returns a two level Toolbox.Workflow.Transition structure. The first level contains parameters important for workflow control. Some parameters are copied from param some are calculated. The second level is located under attributes key. It contains some parameters important for workflow control and all parameters defined by scenario developer. Transition structure is passed, alongside with instance status and message to all callback functions. The scenario developer can evaluate them inside the function. from - The state of the workflow from which the transition starts. required parameter to - The state of the workflow after transition is completed. required paramater description_before - Transition from from state to to state generates an output action. Output action contains simplified list of all possible states that can be reached fron to state. The list is stored in body.attributes[&quot;future&quot;] . Each entry of the list is a map. The description before is assigned to description key of that map. The description intended to provide the human readable description of the conditions leading to the particular state. - required parameter - description after - The aforementioned output action contains a list of all state changes that the workflow has already passed through. That list is stored in body.attributes[&quot;history']. Each entry of the list is a map containing details of the state. The description after is copied to the &quot;description&quot; key of the map. The description is intended to provide human readable description about the past state change. - required parameter - severity - The integer number. The severity of the incident after transition to to state. - optional parameter - when - predicate used to select the transition which will be executed. - optional parameter - there can be multiple when definitions in list, all definitions are evaluated with &amp;&amp; boolean operator - possible when definitions are: - {Module, function} , where function accepts transition, instance and message as args, returns boolean - , where timeout is defined in milliseconds (transition is then automatically executed when time elapses the specified value) - (transition is executed if the specified field of state reaches the specified value) - then - callback used to update workflow instance state during transition execution. - optional paramater - there can be multiple then definitions in list, all definitions are executed in given order - possible then definitions: - {Module, function} , where function accepts transition, instance and message as args, and returns to update the instance state - side effects - callback used to generate side effects during transition execution. - optional parameter - there can be multiple side effect definitions in list, all definitions are executed in given order. - possible definitions: - {Module, function} , where function accepts transition, instance and message as args, and returns - update history_entry - callback used to modify transition execution history entry stored in asset map. - optional parameter - there can be multiple definitions in list, all definitions are executed in given order - this is usually used to interpolate description texts, or to add additional attributes to history - possible definitions: - {Module, function} , where function accepts history entry, transition, instance and message as args and returns - update_possible_transition - callback used to modify possible future transitions stored in asset map. - optional parameter - there can be multiple definitions in list, all definitions are executed in given order - the callback is executed for each possible future transition - note this only modifies the items of future attribute of the incident asset, this has no effect on definition transitions - possible definitions: - {Module, function} , where function accepts possible transition, transition, instance and message as args, and returns - user_actions - specifies all possible user actions from the target state - a map of user actions that should be enabled once the transition is executed and incident is in the target state. - optional parameter - user actions are automatically deleted if not present in the next transition - keys are strings - values are {module, function} , this specifies the function to be called to generate the user action token (since tokens are not known in advance, they are generated by the specified function) - the function takes transition, instance, message as arguments and is expected to return to register the user action token - upsert_attributes - specifies additional attributes to be added to the incident asset. - optional parameter - list of callbacks to compute the additional attributes - each callback produces a map of additional attributes and this is merged into a single map where the latter has priority over the former - attributes cannot override attributes handled by this workflow incl. user actions, only other attributes can be added - possible definitions: - {Module, function} , where function accepts transition, instance and message as args, and should return Any other keys of the params parameter are put into the map under attributes key of the Transition` structure. They can be used inside callback functions for any purpose. For example for debugging: ``` Code snippets: def definition do Incident.new() |&gt; Incident.add_transition( from: &quot;detected&quot;, to: &quot;rare&quot;, when: [{__MODULE , :rare_message_arrived?}], description_before: &quot;Other messages will arrive&quot;, description_after: &quot;Rare message arrived&quot;, update_history_entry: [{__MODULE , :update_history_entry}], my_debug_key : &quot;This should be a very rare transition&quot; ) ... |&gt; Incident.build() end def update_history_entry(history, transition , _workflow_instance, message) do Logger.info(transition.attributes. my_debug_key ) ... {:ok, new_history} end `` When a message is evaluated the callbacks above are run in the following order. 1. when callbacks are evaluated to see if the current transition is ready to be executed. If not the next transition is tried. 2. then callbacks are evaluated to update the instance state. 3. user_actions is evaluated, all callbacks specified inside are executed and all user action tokens are calculated. 4. update_history_entry callbacks are evaluated to update the new history entry 5. update_possible_transition callbacks are evaluated to update the new possible future transitions 6. upsert_attributes callbacks are evaluated to gather additional attributes 7. side_effects callbacks are evaluated to acquire the list of all additional output actions All text bearing attributes (such as subject , name , description_before , description_after ) has access to incident metadata dictionary. This dictionary contains these, keys: - id - incident id - transition - transition attributes dictionary containing from , to , severity keys - state - dictionary containing user defined state - message - altworx message which triggered given transition Metadata can be accessed via interpolation defined as {{}} , e.g. {{state.foo.bar}} , {{message.body.foo}}`.","ref":"Toolbox.Incident.html#add_transition/2"},{"type":"function","title":"Toolbox.Incident.append_create_incident_output_actions/3","doc":"","ref":"Toolbox.Incident.html#append_create_incident_output_actions/3"},{"type":"function","title":"Toolbox.Incident.append_update_incident_output_actions/3","doc":"","ref":"Toolbox.Incident.html#append_update_incident_output_actions/3"},{"type":"function","title":"Toolbox.Incident.build/1","doc":"","ref":"Toolbox.Incident.html#build/1"},{"type":"function","title":"Toolbox.Incident.handle_message/3","doc":"Uses given incident workflow definition and message to update state of given instance. If no configured workflow transition matches, nothing will happen = instance state will remain the same. Order of callback execution: when definitions of transitions in definition order then definitions of matching transition update history entry definitions of matching transition update possible transition definitions of matching transition side effects definitions of matching transition","ref":"Toolbox.Incident.html#handle_message/3"},{"type":"function","title":"Toolbox.Incident.new/0","doc":"Creates new blank incident workflow definition","ref":"Toolbox.Incident.html#new/0"},{"type":"function","title":"Toolbox.Incident.new_instance/7","doc":"Creates new incident instance for given workflow. params can be used to specify additional transition-like parameters. actors parameter is evaluated only when new_instance/7 is called and is not stored in the instance. It is a listing of actors used when an incident is created. Its format is a list of maps where id and type are mandatory keys. id and type references an asset. For other available options see add_transition/2 .","ref":"Toolbox.Incident.html#new_instance/7"},{"type":"function","title":"Toolbox.Incident.update_incident_history_entry/4","doc":"","ref":"Toolbox.Incident.html#update_incident_history_entry/4"},{"type":"function","title":"Toolbox.Incident.update_incident_possible_transition/4","doc":"","ref":"Toolbox.Incident.html#update_incident_possible_transition/4"},{"type":"function","title":"Toolbox.Incident.update_incident_state/3","doc":"","ref":"Toolbox.Incident.html#update_incident_state/3"},{"type":"function","title":"Toolbox.Incident.update_user_actions/3","doc":"","ref":"Toolbox.Incident.html#update_user_actions/3"},{"type":"function","title":"Toolbox.Incident.validate_upsert_attributes/1","doc":"","ref":"Toolbox.Incident.html#validate_upsert_attributes/1"},{"type":"module","title":"Toolbox.Incident.IdGenerator","doc":"Module provides functions to generate incident id.","ref":"Toolbox.Incident.IdGenerator.html"},{"type":"function","title":"Toolbox.Incident.IdGenerator.generate/3","doc":"Generates incident id based on incident type, timestamp and serialization_vector. Serialization vector is list of attributes relevant which together with type and timestamp identifies given incident. iex&gt; Toolbox.Incident.IdGenerator.generate(&quot;/incident/type&quot;, 0123456789, [&quot;/asset/type/1&quot;]) &quot;/incident/type/f4c3877a-d50b-3a16-b6ec-3ebb2b308d40&quot;","ref":"Toolbox.Incident.IdGenerator.html#generate/3"},{"type":"module","title":"Toolbox.Random","doc":"Functions for working with deterministic (pseudo)randomness in scenarios. In some use cases, scenarios need to behave “randomly”, for example, when a scenario serves as a demo events generator. Scenarios are by design deterministic which means that they can only behave randomly in the sense that they make certain decisions “on their own” (without the programmer explicitly choosing the outcome) but every time a scenario run is repeated, the same “random” decisions are made. To make this possible, the seed of the random number generator has to be always initialized to the same value and managed in the way that it is preserved throughout the run. And that is what this module helps with. The module wraps around Erlang :rand module for random number generation. Functions in the :rand module use process dictionary for storing the random number generator state (i.e. the seed). To be able to restore a run, the random number generator state is also kept in the state of a unit in an opaque field called __rand_seed . There are three steps (functions of this module) in order to use randomness in scenarios. First, you have to initialize the unit that uses randomness with init/2 . And then, every time before you want to use randomness, you have to call load/1 with the particular unit. And then store the new seed with save/1 . An example is given below. See the documentation for each of the functions for more detailed information. Example defmodule Scenarios.MyScenario.MyTemplate @behaviour Runbox.Scenario.Template.StageBased alias Toolbox.Random def instances do # singleton unit [ { :unit , &quot;generator&quot; , %{ } } ] end def init ( _start_from , unit ) do unit = Random . init ( unit , 293732 ) { :ok , [ ] , unit } end def handle_message ( msg , unit ) do Random . load ( unit ) ... do something that uses :rand calls ... ( e . g . :rand . uniform / 0 or Enum . random / 1 ) { :noreply , oas , Random . save ( unit ) } end end","ref":"Toolbox.Random.html"},{"type":"macro","title":"Toolbox.Random.init/1","doc":"Shortcut for init/2 called with __MODULE__ as the seed. It is a bit shorter, but on the other hand you have to require Toolbox.Random to be able to use this macro. Beware that the generated values will differ when you rename your template module.","ref":"Toolbox.Random.html#init/1"},{"type":"function","title":"Toolbox.Random.init/2","doc":"Initializes a unit for random number generation. You need to provide an initial seed which may be any term, in particular any integer, string or atom. The value doesn't matter (it only affects the randomly generated values) as long as it is constant. Typically, you would use this function inside the init/2 callback of your template. In case you are using a singleton unit, you may, for example, hardcode an integer here or use __MODULE__ as the initial seed: def init ( _start_from , unit ) do unit = Random . init ( unit , __MODULE__ ) { :ok , [ ] , unit } end If you have multiple units (e.g. unit for each person), you may initialize each unit with a different seed (e.g. the ID of that person): def init ( _start_from , unit ) do unit = Random . init ( unit , unit . attributes . id ) { :ok , [ ] , unit } end This function saves the initial seed into the unit's state under an opaque field called __rand_seed and return the modified unit. Therefore, to use randomness, the unit's state has to be a map. The function also sets the seed in the process dictionary so you can start using randomness right away. But you will probably want to do so later in the handle_message/2 callback.","ref":"Toolbox.Random.html#init/2"},{"type":"function","title":"Toolbox.Random.load/1","doc":"Loads a seed from the unit's state into the process dictionary. Typically, you would call this function in the beginning of a handle_message/2 callback that deals with randomness. It is sufficient to call it once per invocation of the callback. Example def handle_message ( msg , unit ) do Random . load ( unit ) person = Enum . random ( unit . state . people ) room = Enum . random ( unit . state . rooms ) # generate an event that $person went into $room { :noreply , oas , Random . save ( unit ) } end","ref":"Toolbox.Random.html#load/1"},{"type":"function","title":"Toolbox.Random.save/1","doc":"Stores the random generator state in the unit's state. Typically, you would use this function after generating random values (and thus modifying the generator's seed) in the handle_message/2 callback. The function returns the modified unit.","ref":"Toolbox.Random.html#save/1"},{"type":"type","title":"Toolbox.Random.input_seed/0","doc":"","ref":"Toolbox.Random.html#t:input_seed/0"},{"type":"type","title":"Toolbox.Random.seed/0","doc":"","ref":"Toolbox.Random.html#t:seed/0"},{"type":"behaviour","title":"Toolbox.Report","doc":"Behaviour module for notification reports. This behaviour encapsulates notifications which behave like reports = notifications are sent periodically. Example defmodule Scenarios.Template.Reporter do @callback Runbox.Scenario.Template.StageBased alias Crontab.CronExpression alias Scenarios.Template.Reporter.DailyReport alias Toolbox.Message alias Toolbox.Report alias Toolbox.Runtime.Stage.Unit import Crontab.CronExpression @impl true def init ( init_timestamp , % Unit { } = unit ) do # report schedule defined as crontab expression report_schedule = ~e[* * * * *] # timezone used to display localized timestamp in notification timezone = &quot;Europe/Prague&quot; # args to init report state init_arg = :ok { :ok , outputs , report_state } = Report . init ( DailyReport , init_timestamp , report_schedule , timezone , init_arg ) { :ok , outputs , % Unit { unit | state : %{ report_state : report_state } } } end @impl true def handle_message ( % Message { } = msg , % Unit { } = unit ) do { :ok , outputs , report_state } = Report . process_message ( unit . state . report_state , msg ) { :reply , outputs , % Unit { unit | state : %{ unit . state | report_state : report_state } } } end end defmodule Scenarios.Template.Reporter.DailyReport do @callback Toolbox.Report alias Runbox.Scenario.OutputAction alias Toolbox.Message @impl true def init ( :ok ) do { :ok , [ ] , %{ incomming_messages : [ ] } } end @impl true def handle_process_message ( %{ } = state , % Message { } = msg ) do { :ok , [ ] , %{ incomming_messages : [ msg | state . incomming_messages ] } } end @impl true def handle_publish_report ( %{ } = state , publish_timestamp , last_publish_timestamp ) do report_notification = % OutputAction.Notification { type : :daily_report , primary_asset : nil , data : %{ &quot;incomming_messages&quot; =&gt; state . incomming_messages , &quot;publish_time&quot; =&gt; publish_timestamp , &quot;last_publish_time&quot; =&gt; last_timestamp } } { :ok , [ report_notification ] , %{ incomming_messages : [ ] } } end end","ref":"Toolbox.Report.html"},{"type":"callback","title":"Toolbox.Report.handle_process_message/2","doc":"","ref":"Toolbox.Report.html#c:handle_process_message/2"},{"type":"callback","title":"Toolbox.Report.handle_publish_report/3","doc":"","ref":"Toolbox.Report.html#c:handle_publish_report/3"},{"type":"callback","title":"Toolbox.Report.init/1","doc":"","ref":"Toolbox.Report.html#c:init/1"},{"type":"function","title":"Toolbox.Report.init/5","doc":"Initializes report","ref":"Toolbox.Report.html#init/5"},{"type":"function","title":"Toolbox.Report.process_message/2","doc":"Processes report message","ref":"Toolbox.Report.html#process_message/2"},{"type":"function","title":"Toolbox.Report.resolve_ambiguous_datetime/1","doc":"Resolves date time into non-ambiguous DateTime. If AmbiguousDateTime is given, it picks one version. If any other non-ambiguous date time is given it is returned as is.","ref":"Toolbox.Report.html#resolve_ambiguous_datetime/1"},{"type":"type","title":"Toolbox.Report.report_state/0","doc":"","ref":"Toolbox.Report.html#t:report_state/0"},{"type":"type","title":"Toolbox.Report.timestamp/0","doc":"","ref":"Toolbox.Report.html#t:timestamp/0"},{"type":"module","title":"Toolbox.Report.State","doc":"Report state","ref":"Toolbox.Report.State.html"},{"type":"type","title":"Toolbox.Report.State.t/0","doc":"","ref":"Toolbox.Report.State.html#t:t/0"},{"type":"module","title":"Toolbox.Utils.Enum","doc":"A set of utility functions for Enums.","ref":"Toolbox.Utils.Enum.html"},{"type":"function","title":"Toolbox.Utils.Enum.each_while_ok/2","doc":"Returns :ok if fun applied to every element of enumerable returns :ok . Otherwise it returns the first error occurred during fun call. Example iex&gt; each_while_ok ( [ ] , fn _ -&gt; :not_called_at_all end ) :ok iex&gt; each_while_ok ( [ 1 , 2 , 3 ] , fn _ -&gt; :ok end ) :ok iex&gt; each_while_ok ( [ 1 , 2 , 3 ] , &amp; ( if &amp;1 == 2 , do : { :error , &amp;1 } , else : :ok ) ) { :error , 2 }","ref":"Toolbox.Utils.Enum.html#each_while_ok/2"},{"type":"function","title":"Toolbox.Utils.Enum.flatten_levels/2","doc":"Flattens the enum by the specified amount of levels. Flattens only the specified amount of levels, contrary to the standard List.flatten/2 which flattens the list recursively. Handy if you e.g. want to flatten only the first level. Importantly, this operation maintains order of the elements in the list. Examples iex&gt; flatten_levels ( [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ] , 1 ) [ :hello , :team , &quot;this&quot; , &quot;is&quot; , &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] iex&gt; flatten_levels ( [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ] , 2 ) [ :hello , :team , &quot;this&quot; , &quot;is&quot; , &quot;now&quot; , &quot;flat&quot; , [ &quot;list&quot; ] ] iex&gt; flatten_levels ( [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ] , 3 ) [ :hello , :team , &quot;this&quot; , &quot;is&quot; , &quot;now&quot; , &quot;flat&quot; , &quot;list&quot; ] iex&gt; flatten_levels ( [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ] , 4 ) [ :hello , :team , &quot;this&quot; , &quot;is&quot; , &quot;now&quot; , &quot;flat&quot; , &quot;list&quot; ] iex&gt; flatten_levels ( [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ] , 0 ) [ [ :hello , :team ] , [ &quot;this&quot; , &quot;is&quot; ] , [ &quot;now&quot; , [ &quot;flat&quot; , [ &quot;list&quot; ] ] ] ]","ref":"Toolbox.Utils.Enum.html#flatten_levels/2"},{"type":"function","title":"Toolbox.Utils.Enum.map_while_ok/2","doc":"Returns {:ok, list_of_values} if fun applied to every element of enumerable returns {:ok, value} . Otherwise it returns the first error occurred during fun call. Example iex&gt; map_while_ok ( [ ] , fn _ -&gt; :not_called_at_all end ) { :ok , [ ] } iex&gt; map_while_ok ( [ 1 , 2 , 3 ] , &amp; ( { :ok , &amp;1 * 2 } ) ) { :ok , [ 2 , 4 , 6 ] } iex&gt; map_while_ok ( [ 1 , 2 , 3 ] , &amp; ( if &amp;1 == 2 , do : { :error , &amp;1 * 2 } , else : { :ok , &amp;1 * 2 } ) ) { :error , 4 }","ref":"Toolbox.Utils.Enum.html#map_while_ok/2"},{"type":"function","title":"Toolbox.Utils.Enum.reduce_while_ok/3","doc":"Invokes fun for each element in the enumerable with the accumulator. If every fun call returns {:ok, new_acc} , functions returns {:ok, last_acc} . Otherwise it returns the first error occurred during fun call. Example iex&gt; reduce_while_ok ( [ ] , :initial , fn _ , _ -&gt; :not_called_at_all end ) { :ok , :initial } iex&gt; reduce_while_ok ( [ 1 , 2 , 3 ] , 0 , &amp; ( { :ok , &amp;1 + &amp;2 } ) ) { :ok , 6 } iex&gt; reduce_while_ok ( [ 1 , 2 , 3 ] , 0 , &amp; ( if &amp;1 == 2 , do : { :error , &amp;1 + &amp;2 } , else : { :ok , &amp;1 + &amp;2 } ) ) { :error , 3 }","ref":"Toolbox.Utils.Enum.html#reduce_while_ok/3"},{"type":"module","title":"Toolbox.Utils.Map","doc":"A set of utility functions for Maps.","ref":"Toolbox.Utils.Map.html"},{"type":"function","title":"Toolbox.Utils.Map.deep_create/2","doc":"Creates deep map with value on key represented by path . If path is empty, it returns given value . Examples iex&gt; Toolbox.Utils.Map . deep_create ( [ ] , :some_value ) :some_value iex&gt; Toolbox.Utils.Map . deep_create ( [ &quot;a&quot; , :b ] , :other_value ) %{ &quot;a&quot; =&gt; %{ :b =&gt; :other_value } }","ref":"Toolbox.Utils.Map.html#deep_create/2"},{"type":"function","title":"Toolbox.Utils.Map.deep_merge/2","doc":"Deeply updates left map by right map. Existing keys from left map not contained in right map are unchanged. New keys from right map are added. If the key is present in both maps and both values are maps, maps are deeply merges. If the key is present in both maps and value of one of them is not a map, value of key in left map is replaced by value in right map. Example iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : 1 } , %{ b : 2 } ) %{ a : 1 , b : 2 } iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : %{ b : 1 } } , %{ a : %{ c : 3 } } ) %{ a : %{ b : 1 , c : 3 } } iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : %{ b : %{ c : 1 } } } , %{ a : %{ b : %{ d : 2 } } } ) %{ a : %{ b : %{ c : 1 , d : 2 } } } iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : 1 } , %{ a : %{ b : 2 } } ) %{ a : %{ b : 2 } } iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : %{ b : 1 } } , %{ a : 2 } ) %{ a : 2 } iex&gt; Toolbox.Utils.Map . deep_merge ( %{ a : 1 } , %{ a : 2 } ) %{ a : 2 }","ref":"Toolbox.Utils.Map.html#deep_merge/2"},{"type":"function","title":"Toolbox.Utils.Map.deep_remove/2","doc":"Removes key represented as path (list of keys) deeply in map if it exists. Example iex&gt; Toolbox.Utils.Map . deep_remove ( %{ a : 1 , b : 2 } , [ ] ) %{ a : 1 , b : 2 } iex&gt; Toolbox.Utils.Map . deep_remove ( %{ a : 1 , b : 2 } , [ :a ] ) %{ b : 2 } iex&gt; Toolbox.Utils.Map . deep_remove ( %{ a : 1 , b : 2 } , [ :a , :aa ] ) %{ a : 1 , b : 2 } iex&gt; Toolbox.Utils.Map . deep_remove ( %{ a : %{ aa : 1 , ab : 2 } , b : 2 } , [ :a , :aa ] ) %{ a : %{ ab : 2 } , b : 2 } iex&gt; Toolbox.Utils.Map . deep_remove ( %{ a : %{ aa : 1 , ab : 2 } , b : 2 } , [ :a , :aa , :aaa ] ) %{ a : %{ aa : 1 , ab : 2 } , b : 2 }","ref":"Toolbox.Utils.Map.html#deep_remove/2"},{"type":"function","title":"Toolbox.Utils.Map.from_flat_map/1","doc":"Converts map from flat to nested map. The input map must be flat with array keys, which represent the path to the respective values. The corresponding nested map is then constructed. Also see to_flat_map/1 which performs the reverse operation. Examples iex&gt; Toolbox.Utils.Map . from_flat_map ( %{ ...&gt; [ &quot;contact&quot; , &quot;name&quot; ] =&gt; &quot;Jored&quot; , ...&gt; [ &quot;contact&quot; , &quot;address&quot; , &quot;city&quot; ] =&gt; &quot;Lrno&quot; , ...&gt; [ &quot;active&quot; , &quot;ui&quot; ] =&gt; true , ...&gt; [ &quot;active&quot; , &quot;backend&quot; ] =&gt; false ...&gt; } ) %{ &quot;contact&quot; =&gt; %{ &quot;name&quot; =&gt; &quot;Jored&quot; , &quot;address&quot; =&gt; %{ &quot;city&quot; =&gt; &quot;Lrno&quot; } } , &quot;active&quot; =&gt; %{ &quot;ui&quot; =&gt; true , &quot;backend&quot; =&gt; false } } iex&gt; Toolbox.Utils.Map . from_flat_map ( %{ ...&gt; [ :very , :very , :very , :very , :very , :very , :deep ] =&gt; &quot;object&quot; , ...&gt; [ :very , :very , :very , :very , :very , :very , :nested ] =&gt; &quot;map&quot; ...&gt; } ) %{ very : %{ very : %{ very : %{ very : %{ very : %{ very : %{ deep : &quot;object&quot; , nested : &quot;map&quot; } } } } } } } iex&gt; Toolbox.Utils.Map . from_flat_map ( %{ } ) %{ } iex&gt; Toolbox.Utils.Map . from_flat_map ( Toolbox.Utils.Map . to_flat_map ( %{ ...&gt; &quot;contact&quot; =&gt; %{ ...&gt; &quot;name&quot; =&gt; &quot;Jonas&quot; , ...&gt; &quot;address&quot; =&gt; %{ &quot;city&quot; =&gt; &quot;Brnp&quot; } } , ...&gt; &quot;favorites&quot; =&gt; %{ ...&gt; &quot;food&quot; =&gt; %{ ...&gt; 1 =&gt; [ &quot;cornflakes&quot; , &quot;corn&quot; ] , ...&gt; 5 =&gt; [ &quot;pizza&quot; ] , ...&gt; 8 =&gt; [ &quot;pasta&quot; ] ...&gt; } ...&gt; } ...&gt; } ...&gt; ) ) %{ &quot;contact&quot; =&gt; %{ &quot;name&quot; =&gt; &quot;Jonas&quot; , &quot;address&quot; =&gt; %{ &quot;city&quot; =&gt; &quot;Brnp&quot; } } , &quot;favorites&quot; =&gt; %{ &quot;food&quot; =&gt; %{ 1 =&gt; [ &quot;cornflakes&quot; , &quot;corn&quot; ] , 5 =&gt; [ &quot;pizza&quot; ] , 8 =&gt; [ &quot;pasta&quot; ] } } }","ref":"Toolbox.Utils.Map.html#from_flat_map/1"},{"type":"function","title":"Toolbox.Utils.Map.get_path/3","doc":"Follows the path in the (deep) map, defaulting to default if some of the objects on the path is missing. Example iex&gt; object = %{ &quot;some&quot; =&gt; %{ &quot;deep&quot; =&gt; %{ &quot;object&quot; =&gt; &quot;data&quot; } } } ...&gt; Toolbox.Utils.Map . get_path ( object , [ &quot;some&quot; , &quot;deep&quot; , &quot;object&quot; ] ) &quot;data&quot; iex&gt; Toolbox.Utils.Map . get_path ( object , [ &quot;non-existent&quot; ] , &quot;default&quot; ) &quot;default&quot;","ref":"Toolbox.Utils.Map.html#get_path/3"},{"type":"function","title":"Toolbox.Utils.Map.json_paths/1","doc":"Extracts all (json-like) paths in the map. Given any artitrarily nested map, the function returns all paths the map contains. The output is not sorted in a predictable way. Examples iex&gt; Toolbox.Utils.Map . json_paths ( %{ ...&gt; &quot;contact&quot; =&gt; %{ ...&gt; &quot;name&quot; =&gt; &quot;Trest&quot; , ...&gt; &quot;phone&quot; =&gt; 605554171 , ...&gt; &quot;address&quot; =&gt; %{ &quot;city&quot; =&gt; &quot;Brno&quot; } ...&gt; } , ...&gt; &quot;active&quot; =&gt; false ...&gt; } ) |&gt; Enum . sort ( ) [ [ &quot;active&quot; ] , [ &quot;contact&quot; , &quot;address&quot; , &quot;city&quot; ] , [ &quot;contact&quot; , &quot;name&quot; ] , [ &quot;contact&quot; , &quot;phone&quot; ] , ] iex&gt; Toolbox.Utils.Map . json_paths ( %{ &quot;a&quot; =&gt; %{ 1 =&gt; %{ %{ &quot;cool&quot; =&gt; &quot;stuff&quot; } =&gt; %{ atom : :work_too } } } } ) [ [ &quot;a&quot; , 1 , %{ &quot;cool&quot; =&gt; &quot;stuff&quot; } , :atom ] ] iex&gt; Toolbox.Utils.Map . json_paths ( %{ } ) [ ]","ref":"Toolbox.Utils.Map.html#json_paths/1"},{"type":"function","title":"Toolbox.Utils.Map.subtract/2","doc":"Subtracts map from another map, the leaf-paths from the second map are removed from the first. Examples iex&gt; Toolbox.Utils.Map . subtract ( %{ a : 1 , b : 2 } , %{ a : true } ) %{ b : 2 } iex&gt; Toolbox.Utils.Map . subtract ( %{ a : 1 } , %{ a : %{ b : %{ c : true } } } ) %{ a : 1 } iex&gt; Toolbox.Utils.Map . subtract ( ...&gt; %{ a : %{ b : %{ c : &quot;1&quot; } , d : &quot;2&quot; } , e : &quot;3&quot; } , ...&gt; %{ a : %{ b : %{ c : true } } , e : true } ...&gt; ) %{ a : %{ b : %{ } , d : &quot;2&quot; } } iex&gt; Toolbox.Utils.Map . subtract ( ...&gt; %{ &quot;a&quot; =&gt; 1 , &quot;b&quot; =&gt; %{ &quot;c&quot; =&gt; true } } , ...&gt; %{ &quot;b&quot; =&gt; %{ &quot;c&quot; =&gt; true } } ...&gt; ) %{ &quot;a&quot; =&gt; 1 , &quot;b&quot; =&gt; %{ } } iex&gt; Toolbox.Utils.Map . subtract ( %{ a : 1 , b : 2 } , %{ c : 2 , d : 3 } ) %{ a : 1 , b : 2 } iex&gt; Toolbox.Utils.Map . subtract ( %{ a : %{ b : 1 } } , %{ a : true } ) %{ }","ref":"Toolbox.Utils.Map.html#subtract/2"},{"type":"function","title":"Toolbox.Utils.Map.to_flat_map/1","doc":"Converts regular nested map to flat map. The input map can be any map - arbitrarily nested, the output is a flat map where keys are arrays representing the path to the respective value. Also see from_flat_map/1 which performs the reverse operation. Examples iex&gt; Toolbox.Utils.Map . to_flat_map ( %{ ...&gt; &quot;contact&quot; =&gt; %{ ...&gt; &quot;name&quot; =&gt; &quot;Kert&quot; , ...&gt; &quot;address&quot; =&gt; %{ &quot;city&quot; =&gt; &quot;Krno&quot; } ...&gt; } , ...&gt; &quot;status&quot; =&gt; %{ &quot;ui&quot; =&gt; %{ &quot;active&quot; =&gt; true } , ...&gt; &quot;backend&quot; =&gt; %{ &quot;fluent&quot; =&gt; false } } ...&gt; } ) %{ [ &quot;contact&quot; , &quot;name&quot; ] =&gt; &quot;Kert&quot; , [ &quot;contact&quot; , &quot;address&quot; , &quot;city&quot; ] =&gt; &quot;Krno&quot; , [ &quot;status&quot; , &quot;ui&quot; , &quot;active&quot; ] =&gt; true , [ &quot;status&quot; , &quot;backend&quot; , &quot;fluent&quot; ] =&gt; false } iex&gt; Toolbox.Utils.Map . to_flat_map ( %{ good : %{ 1 =&gt; :nice } , very : %{ deep : %{ %{ ish : :map } =&gt; :askey } } } ) %{ [ :good , 1 ] =&gt; :nice , [ :very , :deep , %{ ish : :map } ] =&gt; :askey } iex&gt; Toolbox.Utils.Map . to_flat_map ( %{ &quot;regular&quot; =&gt; &quot;map&quot; } ) %{ [ &quot;regular&quot; ] =&gt; &quot;map&quot; } iex&gt; Toolbox.Utils.Map . to_flat_map ( %{ } ) %{ }","ref":"Toolbox.Utils.Map.html#to_flat_map/1"},{"type":"module","title":"Toolbox.Workflow","doc":"Module provides API to define state machine based workflow. Workflow is represented by struct containing all configuration needed to handle state of individual instances of given workflow. Workflow is defined by transitions, which contains all needed attributes (source node, target node, ..). Workflow defined by transitions needs to be built epxlicitly by calling build/1 function. This function validates all workflow dependencies and state machine structure. Workflow definition produced by build/1 function can be used to create new instance of given workflow. Workflow instance is struct of: unique id status state (map containing all relevant data, manipulated by then callback during status transition) history (record of all transitions which given instance went through in the past) history in default is map of %{&quot;timestamp&quot; =&gt; timestamp, &quot;status&quot; =&gt; binary} history entry can be updated via update_history_entry callback defined in transition possible transitions (list of all transitions which could be executed in close future) possible transition in default is map of %{&quot;timestamp&quot; =&gt; -1 | timestamp, &quot;status&quot; =&gt; binary} timestamp -1 represents non-timeoutable transition possible transition entry can be updated via update_possible_transition callback defined in transition Workflow transition can be defined by: from (source status) to (target status) when (predicate used to select transition which will be executed) there can be multiple when definitions in list, all definitions are connected with &amp;&amp; relation possible when definitions: {Module, function} , where function accepts transition, instance and message as args {:timeout, timeout} , where timeout is defined in milliseconds {:=, [path, to, state, key], value} {:&gt;, [path, to, state, key], value} {:&lt;, [path, to, state, key], value} {:&lt;=, [path, to, state, key], value} {:&gt;=, [path, to, state, key], value} {:contains, [path, to, state, key], value} {:is_in, [path, to, state, key], list_value} all callbacks should return boolean value then (callback used to update workflow instance state during transition execution) there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts transition, instance and message as args all calbacks should return {:ok, %{} = wf_instance_state} | {:error, reason} side_effects (callback used to generate output actions during transition execution) there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts transition, instance and message as args all calbacks should return {:ok, [OutputAction]} | {:error, reason} update_history_entry (callback used to modify transition execution history entry) there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts history entry, transition, instance and message as args all calbacks should return {:ok, %{} = updated_history_entry} | {:error, reason} update_possible_transition (callback used in handle_message/3 to modify possible transition) there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts possible transtion, transition, instance and message as args all calbacks should return {:ok, %{} = updated_possible_transition} | {:error, reason} Workflow transition execution uses given workflow definition and message to update state of given instance. If no configured workflow transition matches, nothing will happend = instance state will remain the same. Order of callback execution: when definitions of transitions in definition order then definitions of matching transition update history entry definitions of matching transition update possible transition definitions of matching transition side effects definitions of matching transition","ref":"Toolbox.Workflow.html"},{"type":"function","title":"Toolbox.Workflow.add_transition/2","doc":"Workflow transition can be defined by keys in params: from (source status) required parameter to (target status) required parameter when (predicate used to select transition which will be executed) optional parameter there can be multiple when definitions in list, all definitions are connected with &amp;&amp; relation possible when definitions: {Module, function} , where function accepts transition, instance and message as args {:timeout, timeout} , where timeout is defined in milliseconds {:=, [path, to, state, key], value} {:&gt;, [path, to, state, key], value} {:&lt;, [path, to, state, key], value} {:&lt;=, [path, to, state, key], value} {:&gt;=, [path, to, state, key], value} {:contains, [path, to, state, key], value} {:is_in, [path, to, state, key], list_value} all callbacks should return boolean value then (callback used to update workflow instance state during transition execution) optional parameter there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts transition, instance and message as args all calbacks should return {:ok, %{} = wf_instance_state} | {:error, reason} side_effects (callback used to generate output actions during transition execution) optional parameter there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts transition, instance and message as args all calbacks should return {:ok, [OutputAction]} | {:error, reason} update_history_entry (callback used to modify transition execution history entry) optional parameter there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts history entry, transition, instance and message as args all calbacks should return {:ok, %{} = updated_history_entry} | {:error, reason} update_possible_transition (callback used in handle_message/3 to modify possible transition) optional parameter there can be multiple then definitions in list, all definitions are executed in given order possible then definitions: {Module, function} , where funciton accepts possible transtion, transition, instance and message as args all calbacks should return {:ok, %{} = updated_possible_transition} | {:error, reason}","ref":"Toolbox.Workflow.html#add_transition/2"},{"type":"function","title":"Toolbox.Workflow.build/1","doc":"Finishes workflow definition, validates all configured dependencies and workflow structure","ref":"Toolbox.Workflow.html#build/1"},{"type":"function","title":"Toolbox.Workflow.handle_message/2","doc":"Uses given workflow definition and message to update state of given instance. If no configured workflow transition matches, nothing will happend = instance state will remain the same. Order of callback execution: when definitions of transitions in definition order then definitions of matching transition update history entry definitions of matching transition update possible transition definitions of matching transition side effects definitions of matching transition","ref":"Toolbox.Workflow.html#handle_message/2"},{"type":"function","title":"Toolbox.Workflow.handle_message/3","doc":"","ref":"Toolbox.Workflow.html#handle_message/3"},{"type":"function","title":"Toolbox.Workflow.new/0","doc":"Creates new blank workflow definition","ref":"Toolbox.Workflow.html#new/0"},{"type":"function","title":"Toolbox.Workflow.new_instance/7","doc":"Creates new instance for given workflow","ref":"Toolbox.Workflow.html#new_instance/7"},{"type":"type","title":"Toolbox.Workflow.status/0","doc":"","ref":"Toolbox.Workflow.html#t:status/0"},{"type":"type","title":"Toolbox.Workflow.t/0","doc":"","ref":"Toolbox.Workflow.html#t:t/0"},{"type":"type","title":"Toolbox.Workflow.transitions/0","doc":"","ref":"Toolbox.Workflow.html#t:transitions/0"},{"type":"module","title":"Toolbox.Workflow.Instance","doc":"Workflow instance is simple struct which keeps state if individual workflow instances. Workflow updates its status/state via defined callbacks during workflow transitions.","ref":"Toolbox.Workflow.Instance.html"},{"type":"type","title":"Toolbox.Workflow.Instance.t/0","doc":"","ref":"Toolbox.Workflow.Instance.html#t:t/0"},{"type":"module","title":"Toolbox.Workflow.Transition","doc":"Transitions forms workflow configuration. Struct is used internally - workflow client should not know internal details about transition struct.","ref":"Toolbox.Workflow.Transition.html"},{"type":"type","title":"Toolbox.Workflow.Transition.builtin_when_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:builtin_when_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.side_effects_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:side_effects_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.t/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:t/0"},{"type":"type","title":"Toolbox.Workflow.Transition.then_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:then_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.update_history_entry_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:update_history_entry_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.update_possible_transition_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:update_possible_transition_fn_def/0"},{"type":"type","title":"Toolbox.Workflow.Transition.when_fn_def/0","doc":"","ref":"Toolbox.Workflow.Transition.html#t:when_fn_def/0"}]